//----------------------------------------------------------------------------
struct Device
{
	struct Font
	{
		GLYPHMETRICSFLOAT	gmf[256];
		UINT				list_base;
		Font() : list_base(0) {}
		virtual ~Font() { if (list_base) glDeleteLists(list_base,256); }
	};
	
	//singletons
	HWND	hwnd;			//parent window
	HDC		dc;				//dev contex
	HGLRC	rc;				//rendering context
	WNDPROC	old_wnd_proc;	//old window procedure
	
	//fonts
	std::map<UINT,Font> fonts;
	int sel_font;	//currently selected font
	int horz_just;	//horizontal justification, 0 - center, 1 - left, 2 - right
	int vert_just;	//vertical justification, 0 - middle, 1 - top, 2 - bottom
	
	BOOL was_resized;			//resize flag
	
	//auto scale
	int as_mode;
	double as_angle;
	double as_min;
	double as_max;
	
	//ctor/dtor
	Device();
	virtual ~Device();
};

Device::Device() : dc(0), rc(0), old_wnd_proc(0), sel_font(0), was_resized(0), as_mode(1),
	as_angle(45), as_min(1), as_max(1000)
{}
Device::~Device()
{
	if (rc) { wglMakeCurrent(0,0); wglDeleteContext(rc); }		//!! dc, rc
	if (dc) ReleaseDC(hwnd,dc);
	if (old_wnd_proc)
	{
		SetWindowLong(hwnd,GWL_WNDPROC,(LONG)old_wnd_proc);
		//show damned child
		HWND damned_child = FindWindowEx(hwnd,0,"mIRC_Status","");
		if (damned_child) ShowWindow(damned_child,SW_SHOW);
		else
		{
			damned_child = FindWindowEx(hwnd,0,"Static","");
			if (damned_child) ShowWindow(damned_child,SW_SHOW);
		}
		InvalidateRect(hwnd,0,true);
	}
}

//------------------------------------------------------------------------------------//
//globals

std::map<HWND,Device> devices;
Device *pdev;
HWND sel_wnd;
char *input, *output;	//GL_TRANSFORM_BIT

//----------------------------------------------------------------------------
//parsing and hashes

int __cdecl compare_hashes(const int *h1, const int *h2)
{
	return *h1 < *h2 ? -1 : *h1 > *h2 ? 1 : 0;
}
int hash_token()
{
	UINT h = 0;
	for (int i = 0; *input && *input != ' ' && *input != '\t'; i++, input++)
	{
		if (*input != '_')
		{
			BYTE a = (*input | 32) - 48 + i;
			h += a;
			h *= h;
		}
	}
	while(*input == ' ' || *input == '\t') input++;
	return h;
}

float* tokens(int number)
{
	if (number < 0 || number > 512) throw("E_INVALID_NUMBER_OF_PARAMETERS");
	static float toks[512];
	for (int i = 0; i < number; i++)
	{	
		float ret_value = 0;
		do
		{
			if ((*input < '0' || *input > '9') && *input != '.' && *input != '-')
			{
				//symbolic constant
				int hash = hash_token();
				int* result = (int*)bsearch(
					&hash,
					arg_hashes,
					sizeof(arg_hashes)/sizeof(int),
					sizeof(int),
					(int (__cdecl *)(const void*,const void*))compare_hashes);
				ret_value += result ? arg_codes[result - arg_hashes] : 0;
			}
			else
			{
				//number
				ret_value += atof(input);
				while (*input && *input != ' ' && *input != '\t') input++;
				while (*input == ' ' || *input == '\t') input++;
			}
			if (*input == '&')
			{
				while (*input && *input != ' ' && *input != '\t') input++;
				while (*input == ' ' || *input == '\t') input++;
			} else break;
		} while(1); 
		toks[i] = ret_value;
	}
	return toks;
}

__forceinline float token() { return *tokens(1); }

//----------------------------------------------------------------------------
template<class T>
void ret_numbers(T *base, int count)
{
	if (count < 0 || count > 512) throw("E_INVALID_NUMBER_OF_ARGUMENTS");
	char buf[1024];
	*output = 0;
	for (int i = 0; i < count; i++)
	{
		sprintf(buf,"%f ",float(base[i]));
		strcat(output,buf);
	}
}

//----------------------------------------------------------------------------
// OpenGL functions

void push_attrib()
{
	UINT f = token();
	glPushAttrib(f ? f : GL_ALL_ATTRIB_BITS);
}
void call_lists()
{
	int num = token();
	token();					//enforce argument type to be GL_FLOAT
	glCallLists(num,GL_FLOAT,tokens(num));
}
void get_boolean()
{
	int type = token();
	int cnt = token();		//how many
	BYTE params[512];
	glGetBooleanv(type,params);
	ret_numbers(params,cnt);
}
void tex_parameter()
{
	float *t = tokens(3);
	glTexParameterf(t[0],t[1],t[2]);
}
void bind_texture()
{
	float *t = tokens(2);
	if (!t[1])
		glBindTexture(GL_TEXTURE_2D,t[0]);
	else
		glBindTexture(t[0],t[1]);
}
void end()
{
	glEnd();
}
void get_material()
{
	int face = token();	//face
	int type = token();	//type
	int cnt = token();	//how many
	float params[512];
	glGetMaterialfv(face,type,params);
	ret_numbers(params,cnt);
}
void cull_face()
{
	glCullFace(token());
}
void rect()
{
	float *t = tokens(4);
	glRectf(t[0],t[1],t[2],t[3]);
}
void depth_range()
{
	float *t = tokens(2);
	glDepthRange(t[0],t[1]);
}
void full_screen()
{
/*	float *f = tokens(2);
	if (f[0] == -2127366080)	//off
	{
		//return to normal mode
		ChangeDisplaySettings(0,0);
	}
	else
	{
		DEVMODE dmSettings;
		ZeroMemory(&dmSettings,sizeof(dmSettings));
		
		if(!EnumDisplaySettings(0,ENUM_CURRENT_SETTINGS,&dmSettings)) throw("E_FAILED");
		dmSettings.dmPelsWidth	= f[0];
		dmSettings.dmPelsHeight	= f[1];
		
		int result = ChangeDisplaySettings(&dmSettings,CDS_FULLSCREEN);	
		if(result != DISP_CHANGE_SUCCESSFUL) throw("E_FAILED");
		return;// SetupPixelFormat(m_dc, depth);
	}
	*/
	throw("E_NOT_ACTUALLY_IMPLEMENTED");
}
void load_identity()
{
	glLoadIdentity();
}
void begin()
{
	glBegin(token());
}
void color_mask()
{
	float *t = tokens(4);
	glColorMask(t[0],t[1],t[2],t[3]);
}
void disk()
{
	float *t = tokens(8);
	GLUquadricObj *quadric = gluNewQuadric();
	gluQuadricDrawStyle(quadric,t[4] ? t[4] : GLU_FILL);
	gluQuadricOrientation(quadric,t[5] ? t[5] : GLU_OUTSIDE);
	gluQuadricNormals(quadric,t[6] ? t[6] : GLU_SMOOTH);
	gluQuadricTexture(quadric,t[7] ? t[7] : GL_FALSE);
	gluDisk(quadric,t[0],t[1],t[2],t[3]);
	gluDeleteQuadric(quadric);
}
void normal()
{
	float *t = tokens(3);
	glNormal3f(t[0],t[1],t[2]);
}
void mult_matrix()
{
	float *t = tokens(16);
	glMultMatrixf(t);
}
void shade_model()
{
	glShadeModel(token());
}
void enable()
{
	glEnable(token());
}
void get_tex_level_parameter()
{
	float ret_value = 0;
	float *t = tokens(3);
	glGetTexLevelParameterfv(t[0],t[1],t[2],&ret_value);
	ret_numbers(&ret_value,1);
}
void copy_pixels()
{
	float *t = tokens(5);
	glCopyPixels(t[0],t[1],t[2],t[3],t[4]);
}
void pixel_transfer()
{
	float *t = tokens(2);
	glPixelTransferf(t[0],t[1]);
}
void polygon_offset()
{
	float *t = tokens(2);
	glPolygonOffset(t[0],t[1]);
}
void get_error()
{
	switch(glGetError())
	{
	case GL_INVALID_ENUM:		throw("E_INVALID_ENUM");
	case GL_INVALID_VALUE:		throw("E_INVALID_VALUE");
	case GL_INVALID_OPERATION:	throw("E_INVALID_OPERATION");
	case GL_STACK_OVERFLOW:		throw("E_STACK_OVERFLOW");
	case GL_STACK_UNDERFLOW:	throw("E_STACK_UNDERFLOW");
	case GL_OUT_OF_MEMORY:		throw("E_OUT_OF_MEMORY");
	}
}
void get_tex_parameter()
{
	int target = token(), pname = token(), cnt = token();
	float params[512];
	glGetTexParameterfv(target,pname,params);
	ret_numbers(params,cnt);
}
void accum()
{
	float *t = tokens(2);
	glAccum(t[0],t[1]);
}
void get_float()
{
	int type = token(), cnt = token();
	float params[512];
	glGetFloatv(type,params);
	ret_numbers(params,cnt);
}
void get_tex_env()
{
	int target = token(), pname = token(), cnt = token();
	float params[512];
	glGetTexEnvfv(target,pname,params);
	ret_numbers(params,cnt);
}
void clip_plane()
{
	double plane[4];
	glClipPlane(token(),plane);
	ret_numbers(plane,4);
}
void material()
{
	int face = token(), pname = token();
	glMaterialfv(face,pname,tokens(4));
}
void matrix_mode()
{
	glMatrixMode(token());
}
void gen_lists()
{
	int base = glGenLists(token());
	ret_numbers(&base,1);
}
void delete_lists()
{
	float *t = tokens(2);
	glDeleteLists(t[0],t[1]);
}
void get_light()
{
	int light = token(), pname = token(), cnt = token();
	float params[512];
	glGetLightfv(light,pname,params);
	ret_numbers(params,cnt);
}
void depth_func()
{
	glDepthFunc(token());
}
void is_enabled()
{
	throw(glIsEnabled(token()) ? "$true" : "$false");
}
void print()
{
	//find font
	std::map<UINT,Device::Font>::iterator i = pdev->fonts.find(pdev->sel_font);
	if (i == pdev->fonts.end()) throw("E_INVALID_FONT");
	Device::Font *pfont = &(i->second);
	
	//align
	float dx = 0, dy = 1e10;
	GLYPHMETRICSFLOAT *p = pfont->gmf;
	int len = strlen(input);
	for (int j = len; j--;)
	{
		dx += p[((BYTE*)input)[j]].gmfCellIncX;
		float h = p[((BYTE*)input)[j]].gmfCellIncY;
		if (dy > h) dy = h;
	}
	if(pdev->horz_just == 1) dx = 0;
	else if (pdev->horz_just != 2) dx *= 0.5;
	if(pdev->vert_just == 1) dy = 0;
	else if (pdev->vert_just != 2) dy *= 0.5;

	//print
	glPushAttrib(GL_LIST_BIT | GL_ENABLE_BIT | GL_TRANSFORM_BIT);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(-dx,-dy,0);
	glEnable(GL_CULL_FACE); glCullFace(GL_BACK);
	glListBase(pfont->list_base);
	glCallLists(len,GL_UNSIGNED_BYTE,input);
	glPopMatrix();
	glPopAttrib();
}
void copy_tex_sub_image()
{
	float *t = tokens(8);
	glCopyTexSubImage2D(t[0],t[1],t[2],t[3],t[4],t[5],t[6],t[7]);
}
void load_texture()
{
	switch(load_bmp(input))
	{
	case LOAD_TEXTUREBMP_COULD_NOT_FIND_OR_READ_FILE:	throw("E_CANNOT_FIND_OR_READ_FILE");
	case LOAD_TEXTUREBMP_ILLEGAL_FILE_FORMAT:			throw("E_ILLEGAL_FILE_FORMAT");
	case LOAD_TEXTUREBMP_OPENGL_ERROR:					throw("E_FAILED");
	case LOAD_TEXTUREBMP_OUT_OF_MEMORY:					throw("E_OUT_OF_MEMORY");
	}
	//glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	//glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
}
void unproject()
{
	double mdl[16], prj[16];
	int vp[4]; 
	glGetDoublev(GL_MODELVIEW_MATRIX,mdl);
	glGetDoublev(GL_PROJECTION_MATRIX,prj);
	glGetIntegerv(GL_VIEWPORT,vp);
	int x = token(), y = token(), z = token();
	double mdl_coords[3];
	gluUnProject(x,y,z,mdl,prj,vp,mdl_coords+0,mdl_coords+1,mdl_coords+2);
	ret_numbers(mdl_coords,3);
}
void push_matrix()
{
	glPushMatrix();
}
void ortho()
{
	float *t = tokens(6);
	glOrtho(t[0],t[1],t[2],t[3],t[4],t[5]);
}
void tex_gen()
{
	int coord = token(), pname = token();
	float *t = tokens(4);
	glTexGenfv(coord,pname,t);
}
void cylinder()
{
	float *t = tokens(9);
	GLUquadricObj *quadric = gluNewQuadric();
	gluQuadricDrawStyle(quadric,t[5] ? t[5] : GLU_FILL);
	gluQuadricOrientation(quadric,t[6] ? t[6] : GLU_OUTSIDE);
	gluQuadricNormals(quadric,t[7] ? t[7] : GLU_SMOOTH);
	gluQuadricTexture(quadric,t[8] ? t[8] : GL_FALSE);
	gluCylinder(quadric,t[0],t[1],t[2],t[3],t[4]);
	gluDeleteQuadric(quadric);
}
void edge_flag()
{
	glEdgeFlag(token());
}
void clear()
{
	int buffers = token();
	glClear(buffers ? buffers : GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}
void color()
{
	float *t = tokens(4);
	glColor4f(t[0],t[1],t[2],t[3]);
}
void bind_font()
{
	pdev->sel_font = token();
	pdev->horz_just = pdev->vert_just = 0;
	while(*input)
	{
		switch(UINT(token()))
		{
		case 1://TOP
			pdev->vert_just = 1;
			break;
		case 2://BOTTOM
			pdev->vert_just = 2;
			break;
		case 1030://LEFT
			pdev->horz_just = 1;
			break;
		case 1031://RIGHT
			pdev->horz_just = 2;
			break;
		default:
			throw("E_INVALID_JUSTIFICATION");
		}
	}
}
void copy_tex_sub_image_2d()
{
	float *t = tokens(8);
	glCopyTexSubImage2D(t[0],t[1],t[2],t[3],t[4],t[5],t[6],t[7]);
}
void scissor()
{
	float *t = tokens(4);
	glScissor(t[0],t[1],t[2],t[3]);
}
void blend_func()
{
	float *t = tokens(2);
	glBlendFunc(t[0],t[1]);
}
void get_integer()
{
	int type = token(), cnt = token();		//how many
	int params[512];
	glGetIntegerv(type,params);
	ret_numbers(params,cnt);
}
void perspective()
{
	float *t = tokens(4);
	gluPerspective(t[0],t[1],t[2],t[3]);
}
void get_tex_gen()
{
	int coord = token(), pname = token(), cnt = token();
	float params[512];
	glGetTexGenfv(coord,pname,params);
	ret_numbers(params,cnt);
}
void line_stipple()
{
	float *t = tokens(2);
	glLineStipple(t[0],t[1]);
}
void gen_textures()
{
	int cnt = token();
	UINT params[512];
	glGenTextures(cnt,params);
	ret_numbers(params,cnt);
}
void clear_accum()
{
	float *t = tokens(4);
	glClearAccum(t[0],t[1],t[2],t[3]);
}
void text_width()
{
	std::map<UINT,Device::Font>::iterator i = pdev->fonts.find(pdev->sel_font);
	if (i == pdev->fonts.end()) throw("E_INVALID_FONT");
	GLYPHMETRICSFLOAT *p = i->second.gmf;
	for (float width = 0; *input;)
		width += p[*(BYTE*)input++].gmfCellIncX;
	ret_numbers(&width,1);
}
void polygon_mode()
{
	float *t = tokens(2);
	glPolygonMode(t[0],t[1]);
}
void version()
{
	throw("nOpenGLn 1.0 by Necroman, #mIRC @ Undernet");	
}
void frustum()
{
	float *t = tokens(6);
	glFrustum(t[0],t[1],t[2],t[3],t[4],t[5]);
}
void is_texture()
{
	glIsTexture(token());	
}
void disable()
{
	glDisable(token());
}
void rotate()
{
	float *t = tokens(4);
	glRotatef(t[0],t[1],t[2],t[3]);
}
void new_list()
{
	float *t = tokens(2);
	glNewList(t[0],t[1] ? t[1] : GL_COMPILE);
}
void detach()
{
	devices.erase(sel_wnd);
}
void fog()
{
	int pname = token(); 
	glFogfv(pname,tokens(4));
}
void pop_attrib()
{
	glPopAttrib();
}
void translate()
{
	float *t = tokens(3);
	glTranslatef(t[0],t[1],t[2]);
}
void pop_matrix()
{
	glPopMatrix();
}
void sphere()	//radius, slices, stacks, draw_style, orientation, normals, texture
{
	float *t = tokens(7);
	GLUquadricObj *quadric = gluNewQuadric();
	gluQuadricDrawStyle(quadric,t[3] ? t[3] : GLU_FILL);
	gluQuadricOrientation(quadric,t[4] ? t[4] : GLU_OUTSIDE);
	gluQuadricNormals(quadric,t[5] ? t[5] : GLU_SMOOTH);
	gluQuadricTexture(quadric,t[6] ? t[6] : GL_FALSE);
	gluSphere(quadric,t[0],t[1],t[2]);
	gluDeleteQuadric(quadric);
}
void delete_font()
{
	pdev->fonts.erase(token());
}
void project()
{
	double mdl[16], prj[16];
	int vp[4]; 
	glGetDoublev(GL_MODELVIEW_MATRIX,mdl);
	glGetDoublev(GL_PROJECTION_MATRIX,prj);
	glGetIntegerv(GL_VIEWPORT,vp);
	double win_coords[3];
	float *t = tokens(3);
	gluProject(t[0],t[1],t[2],mdl,prj,vp,win_coords+0,win_coords+1,win_coords+2);
	ret_numbers(win_coords,3);
}
void create_font()
{

	//initialize LOGFONT structure
	LOGFONT lf;
	ZeroMemory(&lf,sizeof(lf));
	lf.lfOutPrecision = OUT_TT_PRECIS;
	lf.lfQuality = ANTIALIASED_QUALITY;
	
	char face_name[LF_FACESIZE], *p = face_name;
	for(; *input && *input != ' ' && *input != '\t'; p++, input++)
		if (p < face_name + LF_FACESIZE - 1) *p = (*input == '_' ? ' ' : *input);
	while(*input == ' ' || *input == '\t') input++;
	*p = 0;
	strcpy(lf.lfFaceName,*face_name ? face_name : "Arial");
	
	lf.lfWeight = token();
	lf.lfCharSet = token();
	float thickness = token();
	float deviation = token(); if (deviation < 0.005f) deviation = 0.005f;
	lf.lfItalic = token();
	lf.lfUnderline = token();
	lf.lfStrikeOut = token();

	//add new font to dev
	HFONT h_new_font = CreateFontIndirect(&lf);
	if (!h_new_font) throw("E_INVALID_FONT_PARAMETERS");
	HFONT h_old_font = (HFONT)SelectObject(pdev->dc,h_new_font);
	pdev->fonts[pdev->sel_font];
	Device::Font *pfont = &(pdev->fonts.find(pdev->sel_font)->second);
	pfont->list_base = glGenLists(256);
	wglUseFontOutlines(pdev->dc,0,256,pfont->list_base,deviation,thickness,WGL_FONT_POLYGONS,pfont->gmf);			
	SelectObject(pdev->dc,h_old_font);
	DeleteObject(h_new_font);
}
void is_font()
{
	std::map<UINT,Device::Font>::iterator i = pdev->fonts.find(token());
	throw(i == pdev->fonts.end() ? "$false" : "$true");
}
void vertex()
{
	float *t = tokens(3);
	glVertex3f(t[0],t[1],t[2]);
}
void front_face()
{
	glFrontFace(token());	
}
void text_height()
{
	std::map<UINT,Device::Font>::iterator i = pdev->fonts.find(pdev->sel_font);
	if (i == pdev->fonts.end()) throw("E_INVALID_FONT");
	GLYPHMETRICSFLOAT *p = i->second.gmf;
	for (float height = 0; *input;)
	{
		float h = p[*(BYTE*)input++].gmfBlackBoxY;
		if (height < h) height = h;
	}
	ret_numbers(&height,1);
}
void depth_mask()
{
	glDepthMask(token());
}
void end_list()
{
	glEndList();	
}
void is_list()
{
	strcpy(output,glIsList(token()) ? "$true" : "$false");
}
void get_clip_plane()
{
	double plane[4];
	glGetClipPlane(token(),plane);	
	ret_numbers(plane,4);
}
void auto_scale()
{
	pdev->as_mode = token();
	pdev->as_angle = token();
	pdev->as_min = token();
	pdev->as_max = token();
	RECT rect;
	GetClientRect(pdev->hwnd,&rect);
	PostMessage(pdev->hwnd,WM_SIZE,SIZE_RESTORED,MAKELONG(rect.right,rect.bottom));
}
void viewport()
{
	float *t = tokens(4);
	glViewport(t[0],t[1],t[2],t[3]);
}
void tex_coord()
{
	float *t = tokens(2);
	glTexCoord4f(t[0],t[1],0,1);
}
void load_matrix()
{
	glLoadMatrixf(tokens(16));
}
void point_size()
{
	glPointSize(token());
}
void draw_pixels()
{
	int i1 = token(), i2 = token(), i3 = token(), type = token(), cnt = i1*i2;
	type = GL_FLOAT;
	glDrawPixels(i1,i2,i3,type,tokens(cnt));
}
void bitmap()
{
	throw("E_NOT_IMPLEMENTED");
}
void clear_depth()
{
	glClearDepth(token());
}
void tex_env()
{
	float *t = tokens(3);
	glTexEnvf(t[0],t[1],t[2]);
}
void copy_tex_image_2d()
{
	float *t = tokens(8);
	glCopyTexImage2D(t[0],t[1],t[2],t[3],t[4],t[5],t[6],t[7]);
}
void light()
{
	int light = token(), pname = token();
	glLightfv(light,pname,tokens(4));
}
void get_string()
{
	const char *p = const_cast<const char*>((const char*)glGetString(token()));
	throw(p ? p : "E_INVALID_PARAMETER");	
}
void swap_buffers()
{
	SwapBuffers(pdev->dc);
}
void partial_disk()
{
	float *t = tokens(10);
	GLUquadricObj *quadric = gluNewQuadric();
	gluQuadricDrawStyle(quadric,t[6] ? t[6] : GLU_FILL);
	gluQuadricOrientation(quadric,t[7] ? t[7] : GLU_OUTSIDE);
	gluQuadricNormals(quadric,t[8] ? t[8] : GLU_SMOOTH);
	gluQuadricTexture(quadric,t[9] ? t[9] : GL_FALSE);
	gluPartialDisk(quadric,t[0],t[1],t[2],t[3],t[4],t[5]);
	gluDeleteQuadric(quadric);
}
void call_list()
{
	glCallList(token());
}
void pixel_zoom()
{
	float *t = tokens(2);
	glPixelZoom(t[0],t[1]);
}
void scale()
{
	float *t = tokens(3);
	glScalef(t[0],t[1],t[2]);
}
void copy_tex_image()
{
	float *t = tokens(8);
	glCopyTexImage2D(t[0],t[1],t[2],t[3],t[4],t[5],t[6],t[7]);
}
void hint()
{
	float *t = tokens(2);
	glHint(t[0],t[1]);
}
void raster_pos()
{
	float *t = tokens(2);
	glRasterPos2f(t[0],t[1]);
}
void line_width()
{
	glLineWidth(token());
}
void color_material()
{
	float *t = tokens(2);
	glColorMaterial(t[0],t[1]);
}
void alpha_func()
{
	float *t = tokens(2);
	glAlphaFunc(t[0],t[1]);
}
void list_base()
{
	glListBase(token());
}
void clear_color()
{
	float *t = tokens(4);
	glClearColor(t[0],t[1],t[2],t[3]);
}
void delete_textures()
{
	int cnt = token();
	UINT textures[512];
	for (int i = 0; i < cnt; i++)
		textures[i] = token();
	glDeleteTextures(cnt,textures);
}
void look_at()
{
	float *t = tokens(9);
	gluLookAt(t[0],t[1],t[2],t[3],t[4],t[5],t[6],t[7],t[8]);
}
void get_double()
{
	int type = token(), cnt = token();
	double params[512];
	glGetDoublev(type,params);
	ret_numbers(params,cnt);
}
void light_model()
{
	int pname = token(), cnt = token();
	float params[512];
	glLightModelfv(pname,params);
	ret_numbers(params,cnt);
}
